# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Search Features."""
from __future__ import annotations

import datetime as dt
import logging
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import httpx
from azure.common.exceptions import CloudError
from typing_extensions import Self

from ..._version import VERSION
from .azure_data import get_api_headers
from .sentinel_utils import SentinelUtilsMixin, extract_sentinel_response

if TYPE_CHECKING:
    from ...common.timespan import TimeSpan
__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)


class SentinelSearchlistsMixin(SentinelUtilsMixin):
    """Mixin class for Sentinel Watchlist feature integrations."""

    def create_search(  # noqa: PLR0913
        self: Self,
        query: str,
        start: dt.datetime | None = None,
        end: dt.datetime | None = None,
        search_name: str | None = None,
        *,
        timespan: TimeSpan | None = None,
        limit: int = 1000,
    ) -> None:
        """
        Create a Search job.

        Parameters
        ----------
        query : str
            The KQL query to run as a job.
        start : datetime, optional
            The start time for the query, by default 90 days ago.
        end : datetime, optional
            The end time for the query, by default now.
        search_name : str, optional
            A name to apply to the search, by default a random GUID is generated.
        timespan: Timespan, optional
           If defined, overwrite start and end variables.
        limit: int, optional
           Set the maximum number of results to return. Defaults to 1000.

        Raises
        ------
        CloudError
            If there is an error creating the search job.

        """
        if timespan:
            start = timespan.start
            end = timespan.end
        search_end: dt.datetime = end or dt.datetime.now(tz=dt.timezone.utc)
        search_start: dt.datetime = start or (search_end - dt.timedelta(days=90))
        search_name = (search_name or str(uuid4())).replace("_", "")
        search_url: str = (
            self.sent_urls["search"]
            + f"/{search_name}_SRCH?api-version=2021-12-01-preview"
        )
        search_items: dict[str, dict[str, Any]] = {
            "searchResults": {
                "query": f"{query}",
                "limit": limit,
                "startSearchTime": f"{search_start.isoformat()}",
                "endSearchTime": f"{search_end.isoformat()}",
            },
        }
        search_body: dict[str, Any] = extract_sentinel_response(search_items)
        if not self._token:
            err_msg = "Token not found, can't create search."
            raise ValueError(err_msg)
        search_create_response: httpx.Response = httpx.put(
            search_url,
            headers=get_api_headers(self._token),
            json=search_body,
            timeout=60,
        )
        if not search_create_response.is_success:
            raise CloudError(response=search_create_response)
        logger.info("Search job created with for %s_SRCH.", search_name)

    def check_search_status(self: Self, search_name: str) -> bool:
        """
        Check the status of a search job.

        Parameters
        ----------
        search_name : str
            The name of the search job to check.

        Returns
        -------
        bool
            Returns True if search is ready.

        Raises
        ------
        CloudError
            If error in checking the search job status.

        """
        search_name = search_name.strip("_SRCH")
        search_url: str = (
            self.sent_urls["search"]
            + f"/{search_name}_SRCH?api-version=2021-12-01-preview"
        )
        if not self._token:
            err_msg = "Token not found, can't check search status."
            raise ValueError(err_msg)
        search_check_response: httpx.Response = httpx.get(
            search_url,
            headers=get_api_headers(self._token),
        )
        if not search_check_response.is_success:
            raise CloudError(response=search_check_response)

        check_result: str = search_check_response.json()["properties"][
            "provisioningState"
        ]
        logger.info("%s_SRCH status is '%s'", search_name, check_result)
        return check_result == "Succeeded"

    def delete_search(self: Self, search_name: str) -> None:
        """
        Delete a search result.

        Parameters
        ----------
        search_name: str
            The name of the search to delete.

        Raises
        ------
        CloudError
            If an error occurs when attempting to delete the search

        """
        search_name = search_name.strip("_SRCH")
        search_url: str = (
            self.sent_urls["search"]
            + f"/{search_name}_SRCH?api-version=2021-12-01-preview"
        )
        if not self._token:
            err_msg = "Token not found, can't delete search."
            raise ValueError(err_msg)
        search_delete_response: httpx.Response = httpx.delete(
            search_url,
            headers=get_api_headers(self._token),
        )
        if not search_delete_response.is_success:
            raise CloudError(response=search_delete_response)
        logger.info("%s_SRCH set for deletion.", search_name)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
RiskIQ Threat Intelligence Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Callable, ClassVar

import pandas as pd
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyImportExtraError, MsticpyUserError
from ...common.utility import export
from ..lookup_result import LookupStatus
from .ti_provider_base import ResultSeverity, TIPivotProvider, TIProvider

try:
    from passivetotal import analyzer as ptanalyzer
    from passivetotal.analyzer.illuminate.reputation import ReputationScore
    from passivetotal.analyzer.summary import HostnameSummary

    if TYPE_CHECKING:
        import requests
        from passivetotal.analyzer.hostname import Hostname
        from passivetotal.analyzer.ip import IPAddress
        from passivetotal.analyzer.whois import WhoisRecords

except ImportError as imp_err:
    error_msg: str = "Cannot use this feature without passivetotal package installed."
    raise MsticpyImportExtraError(
        error_msg,
        title="Error importing RiskIQ modules.",
        extra="riskiq",
    ) from imp_err

if TYPE_CHECKING:
    from datetime import datetime

    from ...common.timespan import TimeSpan
    from ...init.pivot import Pivot
    from ...init.pivot_core.pivot_register import PivotRegistration

__version__ = VERSION
__author__ = "Mark Kendrick"


@export
class RiskIQ(TIProvider, TIPivotProvider):
    """RiskIQ Threat Intelligence Lookup."""

    _QUERIES: ClassVar[dict[str, str]] = {
        "ipv4": "ALL",
        "ipv4-articles": "articles",
        "ipv4-artifacts": "artifacts",
        "ipv4-certificates": "certificates",
        "ipv4-components": "components",
        "ipv4-cookies": "cookies",
        "ipv4-hostpairchildren": "hostpair_children",
        "ipv4-hostpairparents": "hostpair_parents",
        "ipv4-passivedns": "resolutions",
        "ipv4-projects": "projects",
        "ipv4-malware": "malware",
        "ipv4-rep": "reputation",
        "ipv4-services": "services",
        "ipv4-summary": "summary",
        "ipv4-trackers": "trackers",
        "ipv4-whois": "whois",
        "hostname": "ALL",
        "hostname-articles": "articles",
        "hostname-artifacts": "artifacts",
        "hostname-certificates": "certificates",
        "hostname-components": "components",
        "hostname-cookies": "cookies",
        "hostname-hostpairchildren": "hostpair_children",
        "hostname-hostpairparents": "hostpair_parents",
        "hostname-passivedns": "resolutions",
        "hostname-projects": "projects",
        "hostname-malware": "malware",
        "hostname-rep": "reputation",
        "hostname-summary": "summary",
        "hostname-trackers": "trackers",
        "hostname-whois": "whois",
    }

    # Aliases
    _QUERIES["dns"] = _QUERIES["hostname"]
    _QUERIES["dns-articles"] = _QUERIES["hostname-articles"]
    _QUERIES["dns-artifacts"] = _QUERIES["hostname-artifacts"]
    _QUERIES["dns-certificates"] = _QUERIES["hostname-certificates"]
    _QUERIES["dns-components"] = _QUERIES["hostname-components"]
    _QUERIES["dns-cookies"] = _QUERIES["hostname-cookies"]
    _QUERIES["dns-hostpairchildren"] = _QUERIES["hostname-hostpairchildren"]
    _QUERIES["dns-hostpairparents"] = _QUERIES["hostname-hostpairparents"]
    _QUERIES["dns-passivedns"] = _QUERIES["hostname-passivedns"]
    _QUERIES["dns-projects"] = _QUERIES["hostname-projects"]
    _QUERIES["dns-malware"] = _QUERIES["hostname-malware"]
    _QUERIES["dns-rep"] = _QUERIES["hostname-rep"]
    _QUERIES["dns-summary"] = _QUERIES["hostname-summary"]
    _QUERIES["dns-trackers"] = _QUERIES["hostname-trackers"]
    _QUERIES["dns-whois"] = _QUERIES["hostname-whois"]

    _PIVOT_ENTITIES: ClassVar[dict[str, dict[str, str]]] = {
        "services": {"IpAddress": "Address"},
        **{
            prop: {"Dns": "DomainName", "IpAddress": "Address", "Host": "fqdn"}
            for prop in [
                "articles",
                "artifacts",
                "certificates",
                "components",
                "cookies",
                "hostpair_children",
                "hostpair_parents",
                "resolutions",
                "projects",
                "malware",
                "reputation",
                "summary",
                "trackers",
                "whois",
            ]
        },
    }

    _REFERENCE: ClassVar[str] = "https://community.riskiq.com"

    def __init__(
        self: RiskIQ,
        *,
        ApiID: str | None = None,  # noqa:N803
        AuthKey: str | None = None,  # noqa:N803
    ) -> None:
        """Instantiate RiskIQ class."""
        super().__init__()
        ptanalyzer.init(username=ApiID, api_key=AuthKey)
        self._pivot_timespan_start: datetime | None = None
        self._pivot_timespan_end: datetime | None = None
        self._pivot_get_timespan: Callable[..., TimeSpan] | None = None

    @property
    def _httpx_client(self: Self) -> requests.Session:
        """Return the PT Analyzer session."""
        return ptanalyzer.api_clients["Cards"].session

    @_httpx_client.setter
    def _httpx_client(self: Self, session: requests.Session) -> None:
        """Set the PT Analyzer session."""
        # pylint: disable=consider-using-dict-items
        for name in ptanalyzer.api_clients:
            ptanalyzer.api_clients[name].session = session

    @staticmethod
    def _severity_rep(classification: str) -> ResultSeverity:
        """Get the severity level for a reputation score classification."""
        return {
            "MALICIOUS": ResultSeverity.high,
            "SUSPICIOUS": ResultSeverity.warning,
            "UNKNOWN": ResultSeverity.information,
            "GOOD": ResultSeverity.information,
        }.get(classification, ResultSeverity.information)

    def lookup_ioc(
        self: Self,
        ioc: str,
        ioc_type: str | None = None,
        query_type: str | None = None,
        *,
        provider_name: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.
        provider_name : str, optional
            Name of the provider to use for query

        Returns
        -------
        pd.DataFrame
            The returned results.

        """
        result: dict[str, Any] = self._check_ioc_type(
            ioc=ioc,
            ioc_type=ioc_type,
            query_subtype=query_type,
        )

        if result["Status"]:
            return pd.DataFrame([result])

        result["Provider"] = provider_name or self.__class__.__name__
        result["Reference"] = self._REFERENCE

        if query_type is None:
            prop = "ALL"
        elif query_type not in [q.split("-", maxsplit=1)[-1] for q in self._QUERIES]:
            result["Result"] = False
            result["Status"] = LookupStatus.QUERY_FAILED.value
            result["Details"] = f"ERROR: unsupported query type {query_type}"
            return pd.DataFrame([result])
        else:
            prop = self._QUERIES.get(f"{result['IocType']}-{query_type}", "ALL")

        try:
            ptanalyzer.set_context("msticpy", "ti", VERSION, prop)
            pt_obj: IPAddress | Hostname = ptanalyzer.get_object(ioc)
            if prop == "ALL":
                result = self._parse_result_all_props(pt_obj, result)
            else:
                result = self._parse_result_prop(pt_obj, prop, result)
        except ptanalyzer.AnalyzerError as err:
            result["Result"] = False
            result["Status"] = LookupStatus.QUERY_FAILED.value
            result["Details"] = f"ERROR: {err}"
            result["Raw_result"] = err
            result["Severity"] = ResultSeverity.unknown.name

        return pd.DataFrame([result])

    def _parse_result_all_props(
        self: Self,
        pt_result: IPAddress | Hostname,
        ti_result: dict[str, Any],
    ) -> dict[str, Any]:
        """Parse results for ALL properties."""
        ti_result["Details"] = {
            "summary": pt_result.summary.as_dict,
            "reputation": pt_result.reputation.as_dict,
        }
        ti_result["RawResult"] = ti_result["Details"]
        ti_result["Result"] = (
            pt_result.summary.total != 0 or pt_result.reputation.score != 0
        )

        rep_severity: ResultSeverity = self._severity_rep(
            pt_result.reputation.classification,
        )
        ti_result["Severity"] = rep_severity.name
        if (
            "malware_hashes" in pt_result.summary.available
            or "articles" in pt_result.summary.available
        ):
            ti_result["Severity"] = (max(rep_severity, ResultSeverity.high)).name
        elif "projects" in pt_result.summary.available:
            ti_result["Severity"] = (max(rep_severity, ResultSeverity.warning)).name
        return ti_result

    def _parse_result_prop(
        self: Self,
        pt_result: IPAddress | Hostname,
        pt_prop: str,
        ti_result: dict[str, Any],
    ) -> dict[str, Any]:
        """Parse result for a specific property."""
        attr = getattr(pt_result, pt_prop)
        if pt_prop == "reputation":
            ti_result["Severity"] = self._severity_rep(attr.classification).name
        elif pt_prop == "malware_hashes" and len(attr) > 0:
            ti_result["Severity"] = ResultSeverity.high.name
        else:
            ti_result["Severity"] = ResultSeverity.information.name
        ti_result["Details"] = ti_result["RawResult"] = attr.as_dict
        ti_result["Result"] = True
        return ti_result

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        del response
        return (True, ResultSeverity.information, None)

    def _set_pivot_timespan(
        self: Self,
        *,
        start: datetime | None = None,
        end: datetime | None = None,
    ) -> bool:
        """
        Set the pivot timespan and track whether it has changed.

        Returns
        -------
        bool
            whether the timespan changed.

        """
        changed = False
        start = (
            start or self._pivot_get_timespan().start
            if self._pivot_get_timespan
            else None
        )
        end = (
            end or self._pivot_get_timespan().end if self._pivot_get_timespan else None
        )
        if (
            start
            and end
            and (start != self._pivot_timespan_start or end != self._pivot_timespan_end)
        ):
            changed = True
            self._pivot_timespan_start = start
            self._pivot_timespan_end = end
            ptanalyzer.set_date_range(start_date=start, end_date=end)
        return changed

    def pivot_value(  # pylint: disable=too-many-arguments #noqa:PLR0913
        self: Self,
        prop: str,
        host: str,
        *,
        start: datetime | None = None,
        end: datetime | None = None,
        exclude_links: bool = True,
        explode_rules: bool = False,
        drop_links: bool = False,
        **kwargs: str,
    ) -> pd.DataFrame:
        """Perform a pivot on a single value."""
        ts_changed: bool = self._set_pivot_timespan(start=start, end=end)
        ptanalyzer.set_context("msticpy", "pivot", VERSION, prop)
        obj: IPAddress | Hostname = ptanalyzer.get_object(host)
        if ts_changed and prop not in ["reputation", "summary", "whois"]:
            obj.reset(prop)
        try:
            attrib: HostnameSummary | WhoisRecords | ReputationScore = getattr(
                obj,
                prop,
            )
        except ptanalyzer.AnalyzerAPIError as err:
            raise RiskIQAPIUserError(err) from err
        except ptanalyzer.AnalyzerError as err:
            err_msg: str = "Analyzer error."
            raise RiskIQUserError(err_msg) from err
        if isinstance(attrib, HostnameSummary):
            return attrib.to_dataframe(exclude_links=exclude_links)
        if isinstance(attrib, ReputationScore):
            return attrib.to_dataframe(
                explode_rules=explode_rules,
                drop_links=drop_links,
            )
        return attrib.to_dataframe(**kwargs)

    def register_pivots(
        self: Self,
        pivot_reg: type[PivotRegistration],
        pivot: Pivot,
    ) -> None:
        """
        Register pivot functions for the TI Provider.

        Parameters
        ----------
        pivot_reg : PivotRegistration
            Pivot registration settings.
        pivot : Pivot
            Pivot library instance

        """
        self._pivot_get_timespan = pivot.get_timespan
        self._pivot_timespan_start = None
        self._pivot_timespan_end = None
        for prop, entity_map in self._PIVOT_ENTITIES.items():
            reg = pivot_reg(
                func_new_name=prop,
                func_static_params={"prop": prop},
                input_type="value",
                entity_map=entity_map,
                entity_container_name="RiskIQ",
                func_df_param_name="data",
                func_df_col_param_name="host",
                func_input_value_arg="host",
                func_out_column_name="query",
            )
            fun = partial(self.pivot_value)
            fun.__doc__ = getattr(
                ptanalyzer.Hostname,
                prop,
                getattr(ptanalyzer.IPAddress, prop),
            ).__doc__
            pivot.add_pivot_function(fun, pivot_reg=reg, container="RiskIQ")


class RiskIQUserError(MsticpyUserError):
    """Generic RiskIQ provider exception."""

    def __init__(
        self: RiskIQUserError,
        message: str,
        help_uri: tuple[str, str] | str | None = None,
    ) -> None:
        """
        Create RiskIQ provider exception.

        Parameters
        ----------
        message: str
            Error message
        help_uri : Union[Tuple[str, str], str, None], optional
            Override the default help URI.

        """
        uri = help_uri or self.DEF_HELP_URI
        super().__init__(
            message,
            help_uri=uri,
            title="error using RiskIQ python library",
            ptlib_uri=(
                "RiskIQ PassiveTotal Python Library",
                "https://passivetotal.readthedocs.io",
            ),
            riqinfo_uri=("RiskIQ Support", "https://www.riskiq.com/resources/support/"),
        )


class RiskIQAPIUserError(RiskIQUserError):
    """RiskIQ API provider exception."""

    def __init__(
        self: RiskIQAPIUserError,
        api_exception: ptanalyzer.AnalyzerAPIError,
    ) -> None:
        """
        Create RiskIQ API exception.

        Parameters
        ----------
        api_exception : ptanalyzer.AnalyzerAPIError
            Underlying API exception.

        """
        title: str = f"{api_exception.status_code} {api_exception.message}"
        super().__init__(title, str(api_exception))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Get proxy settings from config.

Settings format

.. code-block:: yaml

    msticpy:
        Proxies:
            http:
                Url: proxy_url
                UserName: user
                Password: [PLACEHOLDER]
            https:
                Url: proxy_url
                UserName: user
                Password: [PLACEHOLDER]

The entries for the username and password can be specified
as a string, an dictionary of EnvironmentVar: ENV_VAR_NAME
or a dictionary of one of the following:
- KeyVault: None
- KeyVault: secret_name
- KeyVault: vault_name/secret_name

"""
from typing import Dict, Optional

from .pkg_config import get_config
from .provider_settings import get_protected_setting


def get_http_proxies() -> Optional[Dict[str, str]]:
    """Return proxy settings from config."""
    proxy_config = get_config("msticpy.Proxies", None)
    if not proxy_config:
        return None
    proxy_dict = {}
    for protocol, settings in proxy_config.items():
        if "Url" not in settings:
            continue
        if "//" in settings["Url"]:
            prot_str, host = settings["Url"].split("//", maxsplit=1)
        else:
            prot_str, host = protocol, settings["Url"]
        user = (
            get_protected_setting("http.proxies.{protocol}", "UserName")
            if "UserName" in settings
            else None
        )
        password = (
            get_protected_setting("http.proxies.{protocol}", "Password")
            if "Password" in settings
            else None
        )
        auth_str = f"{user}:{password}@" if user else ""

        proxy_dict[protocol.casefold()] = f"{prot_str}//{auth_str}{host}"
    return proxy_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Watchlist Features."""
from __future__ import annotations

import logging
from typing import Any
from uuid import uuid4

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import (
    SentinelUtilsMixin,
    extract_sentinel_response,
    get_http_timeout,
)

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)


class SentinelWatchlistsMixin(SentinelUtilsMixin):
    """Mixin class for Sentinel Watchlist feature integrations."""

    def list_watchlists(self: Self) -> pd.DataFrame:
        """
        List Deployed Watchlists.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the watchlists

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        return self._list_items(
            item_type="watchlists",
            api_version="2021-04-01",
        )

    def create_watchlist(  # noqa: PLR0913
        self: Self,
        watchlist_name: str,
        description: str,
        search_key: str,
        provider: str = "MSTICPy",
        source: str = "Notebook",
        data: pd.DataFrame | None = None,
    ) -> str | None:
        """
        Create a new watchlist.

        Parameters
        ----------
        watchlist_name : str
            The name of the watchlist you want to create,
            this can't be the name of an existing watchlist.
        description : str
            A description of the watchlist to be created.
        search_key : str
            The search key is used to optimize query performance when using watchlists
            for joins with other data.
            This should be the key column that will be used in the watchlist when joining
            to other data tables.
        provider : str, optional
            This is the label attached to the watchlist showing who created it, by default "MSTICPy"
        source : str, optional
            The source of the data to be put in the watchlist, by default "Notebook"
        data: pd.DataFrame, optional
            The data you want to upload to the watchlist

        Returns
        -------
        Optional[str]
            The name/ID of the watchlist.

        Raises
        ------
        MsticpyUserError
            Raised if the watchlist name already exists.
        CloudError
            If there is an issue creating the watchlist.

        """
        self.check_connected()
        if self._check_watchlist_exists(watchlist_name):
            err_msg: str = f"Watchlist {watchlist_name} already exist."
            raise MsticpyUserError(err_msg)
        watchlist_url: str = self.sent_urls["watchlists"] + f"/{watchlist_name}"
        params: dict[str, str] = {"api-version": "2021-04-01"}
        data_items: dict[str, str] = {
            "displayName": watchlist_name,
            "source": source,
            "provider": provider,
            "description": description,
            "itemsSearchKey": search_key,
            "contentType": "text/csv",
        }
        if isinstance(data, pd.DataFrame) and not data.empty:
            data_items["rawContent"] = str(data.to_csv(index=False))
        request_data: dict[str, Any] = extract_sentinel_response(data_items, props=True)
        if not self._token:
            err_msg = "Token not found, can't create watchlist."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            watchlist_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(request_data),
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)

        logger.info("Watchlist created.")
        return response.json().get("name")

    def list_watchlist_items(
        self: Self,
        watchlist_name: str,
    ) -> pd.DataFrame:
        """
        List items in a watchlist.

        Parameters
        ----------
        watchlist_name : str
            The name of the watchlist to get items from

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the watchlists

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        watchlist_name_str: str = f"/{watchlist_name}/watchlistItems"
        return self._list_items(
            item_type="watchlists",
            api_version="2021-04-01",
            appendix=watchlist_name_str,
        )

    def add_watchlist_item(
        self: Self,
        watchlist_name: str,
        item: dict | pd.Series | pd.DataFrame,
        *,
        overwrite: bool = False,
    ) -> None:
        """
        Add or update an item in a Watchlist.

        Parameters
        ----------
        watchlist_name : str
            The name of the watchlist to add items to
        item : Union[Dict, pd.Series, pd.DataFrame]
            The item to add, this can be a dictionary of valies, a Pandas Series, or DataFrame
        overwrite : bool, optional
            Wether you want to overwrite an item if it already exists in the watchlist,
            by default False

        Raises
        ------
        MsticpyUserError
            If the specified Watchlist does not exist.
        MsticpyUserError
            If the item already exists in the Watchlist and overwrite is set to False
        CloudError
            If the API returns an error.

        """
        self.check_connected()
        # Check requested watchlist actually exists
        if not self._check_watchlist_exists(watchlist_name):
            err_msg: str = f"Watchlist {watchlist_name} does not exist."
            raise MsticpyUserError(err_msg)

        new_items: list[dict] = []
        # Convert items to add to dictionary format
        if isinstance(item, pd.Series):
            new_items = [dict(item)]
        elif isinstance(item, dict):
            new_items = [item]
        elif isinstance(item, pd.DataFrame):
            for _, line_item in item.iterrows():
                new_items.append(dict(line_item))

        current_items: pd.DataFrame = self.list_watchlist_items(watchlist_name)
        current_items_values: pd.DataFrame = current_items.filter(
            regex="^properties.itemsKeyValue.",
            axis=1,
        )
        current_items_values.columns = current_items_values.columns.str.replace(
            "properties.itemsKeyValue.",
            "",
            regex=False,
        )

        for new_item in new_items:
            # See if item already exists, if it does get the item ID
            current_df, item_series = current_items_values.align(
                pd.Series(new_item),
                axis=1,
                copy=False,
            )
            if (current_df == item_series).all(axis=1).any() and overwrite:
                watchlist_id: str = current_items[
                    current_items.isin(list(new_item.values())).any(axis=1)
                ]["properties.watchlistItemId"].iloc[0]
            # If not in watchlist already generate new ID
            elif not (current_df == item_series).all(axis=1).any():
                watchlist_id = str(uuid4())
            else:
                err_msg = "Item already exists in the watchlist. Set overwrite = True to replace."
                raise MsticpyUserError(err_msg)

            watchlist_url: str = (
                f"{self.sent_urls['watchlists']}/{watchlist_name}"
                f"/watchlistItems/{watchlist_id}"
            )
            if not self._token:
                err_msg = "Token not found, can't add watchlist item."
                raise ValueError(err_msg)
            response: httpx.Response = httpx.put(
                watchlist_url,
                headers=get_api_headers(self._token),
                params={"api-version": "2021-04-01"},
                content=str({"properties": {"itemsKeyValue": item}}),
                timeout=get_http_timeout(),
            )
            if not response.is_success:
                raise CloudError(response=response)

        logger.info("Items added to %s", watchlist_name)

    def delete_watchlist(
        self: Self,
        watchlist_name: str,
    ) -> None:
        """
        Delete a selected Watchlist.

        Parameters
        ----------
        watchlist_name : str
            The name of the Watchlist to deleted

        Raises
        ------
        MsticpyUserError
            If Watchlist does not exist.
        CloudError
            If the API returns an error.

        """
        self.check_connected()
        # Check requested watchlist actually exists
        if not self._check_watchlist_exists(watchlist_name):
            err_msg: str = f"Watchlist {watchlist_name} does not exist."
            raise MsticpyUserError(err_msg)
        watchlist_url: str = self.sent_urls["watchlists"] + f"/{watchlist_name}"
        params: dict[str, str] = {"api-version": "2021-04-01"}
        if not self._token:
            err_msg = "Token not found, can't delete watchlist."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.delete(
            watchlist_url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        logger.info("Watchlist %s deleted", watchlist_name)

    def delete_watchlist_item(
        self: Self,
        watchlist_name: str,
        watchlist_item_id: str,
    ) -> None:
        """
        Delete a Watchlist item.

        Parameters
        ----------
        watchlist_name : str
            The name of the watchlist with the item to be deleted
        watchlist_item_id : str
            The watchlist item ID to delete

        Raises
        ------
        MsticpyUserError
            If the specified Watchlist does not exist.
        CloudError
            If the API returns an error.

        """
        self.check_connected()
        # Check requested watchlist actually exists
        if not self._check_watchlist_exists(watchlist_name):
            err_msg: str = f"Watchlist {watchlist_name} does not exist."
            raise MsticpyUserError(err_msg)

        watchlist_url: str = (
            self.sent_urls["watchlists"]
            + f"/{watchlist_name}/watchlistItems/{watchlist_item_id}"
        )
        if not self._token:
            err_msg = "Token not found, can't delete watchlist item."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.delete(
            watchlist_url,
            headers=get_api_headers(self._token),
            params={"api-version": "2023-02-01"},
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)

        logger.info("Item deleted from %s", watchlist_name)

    def _check_watchlist_exists(
        self: Self,
        watchlist_name: str,
    ) -> bool:
        """
        Check whether a Watchlist exists or not.

        Parameters
        ----------
        watchlist_name : str
            The Watchlist to check for.
        res_id : str, optional
            The Resource ID of the Sentinel workspace to check in, by default None

        Returns
        -------
        bool
            Whether the Watchlist exists or not.

        """
        # Check requested watchlist actually exists
        existing_watchlists: pd.DataFrame = self.list_watchlists()
        if existing_watchlists.empty:
            return False
        return watchlist_name in existing_watchlists["name"].to_numpy()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""

import contextlib
import json
import random
from json import JSONDecodeError
from typing import Any, Callable, List, Optional, Tuple

import ipywidgets as widgets
import pandas as pd
from deprecated.sphinx import deprecated
from IPython.display import HTML, display
from ipywidgets import Layout

from .._version import VERSION
from .core import IPyDisplayMixin

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-instance-attributes
class SelectAlert(IPyDisplayMixin):
    """
    Alert Selector.

    View list of alerts and select one for investigation.
    Optionally provide and action to call with the selected alert as a parameter
    (typically used to display the alert.)

    Attributes
    ----------
    selected_alert : SecurityAlert
        The selected alert
    alert_id : str
        The SystemAlertId of the selected alert
    alerts : List[SecurityAlert]
        The current alert list (DataFrame)
    action : Callable[..., Optional[Tuple[...]]]
        The callback action to execute on selection
        of an alert.

    """

    _ALERTID_REGEX = r"\[id:(?P<alert_id>.*)\]$"

    def __init__(
        self,
        alerts: pd.DataFrame,
        action: Callable[..., Optional[Tuple]] = None,
        columns: List[str] = None,
        auto_display: bool = False,
        id_col: str = "SystemAlertId",
        **kwargs,
    ):
        """
        Create a new instance of AlertSelector.

        Parameters
        ----------
        alerts : pd.DataFrame
            DataFrame of alerts.
        action : Callable[..., Optional[Tuple]], optional
            Optional function to execute for each selected alert.
            If the function returns one or a tuple of displayable objects
            these will be displayed.
        columns : List[str], optional
            Override the default column names to use from `alerts` to
            generate the select list item summary
            (the default is ['AlertName', 'ProductName'])
        time_col : str, optional
            The column in your alerts that determines when it was created
            Default is 'StartTimeUtc' with a fallback to 'TimeGenerated'.
        id_col : str, optional
            The column in your data that determines the alert id
            Default is 'SystemAlertId'.
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)
        default_alert: str, optional
            If you want to select a default value provide the ID.

        """
        self.alerts = alerts
        self.alert_action = action
        self.id_col = id_col
        time_col = kwargs.get("time_col", "TimeGenerated")
        self.time_col = time_col if time_col in alerts.columns else "TimeGenerated"
        self.default_alert = kwargs.get("default_alert", None)

        columns = columns or ["AlertName", "ProductName"]
        self.disp_columns = list({col for col in columns if col in alerts.columns})
        if not self.disp_columns:
            raise ValueError(
                f"Display columns {','.join(columns)} not found in alerts."
            )
        self._select_items = self._get_select_options(
            alerts, self.time_col, self.id_col, self.disp_columns
        )

        self.selected_alert = None
        self.alert_id = None

        self._w_select_alert = widgets.Select(
            options=self._select_items,
            description="Select alert :",
            layout=Layout(width="95%", height="300px"),
            style={"description_width": "initial"},
            value=self.default_alert,
        )
        self._w_filter_alerts = widgets.Text(
            description="Filter alerts by title:",
            style={"description_width": "initial"},
        )

        # setup to use updatable display objects
        rand_id = random.randint(0, 999999)  # nosec
        self._output_id = f"{self.__class__.__name__}_{rand_id}"
        self._disp_elems: List[Any] = []

        # set up observer callbacks
        self._w_filter_alerts.observe(self._update_options, names="value")
        self._w_select_alert.observe(self._select_alert, names="value")
        wgt_list = [self._w_filter_alerts, self._w_select_alert]
        self._w_display_details = widgets.Checkbox(
            value=True,
            description="Display details",
        )
        if action:
            self._w_display_details.observe(self._run_action, names="value")
            wgt_list.append(self._w_display_details)
        self.layout = widgets.VBox(wgt_list)

        if auto_display:
            self.display()

    @property
    def value(self):
        """Return selected alert."""
        return self.selected_alert

    def display(self):
        """Display the interactive widgets."""
        super().display()
        self._select_top_alert()

    @staticmethod
    def _alert_summary(alert_row, time_col, id_col, columns):
        """Return summarized string of alert properties."""
        items = [f"{alert_row[time_col]}"]
        items.extend(alert_row[col] for col in columns)
        items.append(alert_row[id_col])
        return " || ".join(items)

    @staticmethod
    def _get_select_options(data, time_col, id_col, columns):
        """Return dictionary of displayable options."""
        data_cols = list({time_col, id_col, *columns})
        return {
            val: key
            for key, val in data[data_cols]
            .sort_values(time_col, ascending=True)
            .set_index(id_col)
            .apply(lambda x: " - ".join(str(c) for c in x), axis=1)
            .to_dict()
            .items()
        }

    def _update_options(self, change):
        """Filter the alert list by substring."""
        if change is not None and "new" in change:
            self._w_select_alert.options = {
                alert_desc: alert_id
                for alert_desc, alert_id in self._select_items.items()
                if change["new"].casefold() in alert_desc.casefold()
            }

    def _select_alert(self, selection=None):
        """Select action triggered by picking item from list."""
        if (
            selection is None
            or "new" not in selection
            or not isinstance(selection["new"], str)
        ):
            self.selected_alert = None
        else:
            self.alert_id = selection.new

            self.selected_alert = self._get_alert(self.alert_id)
            if self.alert_action is not None:
                self._run_action()

    def _get_alert(self, alert_id):
        """Get the alert by alert_id."""
        self.alert_id = alert_id
        selected_alerts = self.alerts[self.alerts[self.id_col] == alert_id]
        if selected_alerts.shape[0] > 0:
            alert = pd.Series(selected_alerts.iloc[0])
            if "ExtendedProperties" in alert.index and isinstance(
                alert["ExtendedProperties"], str
            ):
                with contextlib.suppress(JSONDecodeError):
                    alert["ExtendedProperties"] = json.loads(
                        (alert["ExtendedProperties"])
                    )
            if "Entities" in alert.index and isinstance(alert["Entities"], str):
                with contextlib.suppress(JSONDecodeError):
                    alert["Entities"] = json.loads((alert["Entities"]))
            return alert
        return None

    def _select_top_alert(self):
        """Select the first alert by default."""
        top_alert = self.alerts.iloc[0]
        if self.default_alert:
            top_alert = self.alerts[
                self.alerts[self.id_col] == self.default_alert
            ].iloc[0]
        if not top_alert.empty:
            self._w_select_alert.index = 0
            self.alert_id = top_alert[self.id_col]
            self.selected_alert = self._get_alert(self.alert_id)
            if self.alert_action is not None:
                self._run_action()

    def _run_action(self, change=None):
        """Run any action function and display details, if any."""
        del change
        output_objs = None
        if self._w_display_details.value:
            output_objs = self.alert_action(self.selected_alert)
        if output_objs is None:
            self._clear_display()
            return
        if not isinstance(output_objs, (tuple, list)):
            output_objs = [output_objs]
        display_objs = bool(self._disp_elems)
        for idx, out_obj in enumerate(output_objs):
            if not display_objs:
                self._disp_elems.append(
                    display(out_obj, display_id=f"{self._output_id}_{idx}")
                )
            elif idx == len(self._disp_elems):
                break
            else:
                self._disp_elems[idx].update(out_obj)

    def _clear_display(self):
        """Clear any current details."""
        if not self._disp_elems:
            return
        for disp_obj in self._disp_elems:
            disp_obj.update(HTML(""))


# pylint: disable=too-many-instance-attributes
@deprecated(
    reason="Superceded by SelectAlert. Will be removed in v2.0.0.", version="0.5.2"
)
class AlertSelector(SelectAlert):
    """
    AlertSelector.

    View list of alerts and select one for investigation.
    Optionally provide and action to call with the selected alert as a parameter
    (typically used to display the alert.)

    Attributes
    ----------
    selected_alert : SecurityAlert
        The selected alert
    alert_id : str
        The SystemAlertId of the selected alert
    alerts : List[SecurityAlert]
        The current alert list (DataFrame)
    action : Callable[..., None]
        The callback action to execute on selection
        of an alert.

    """

    def __init__(
        self,
        alerts: pd.DataFrame,
        action: Callable[..., None] = None,
        columns: List[str] = None,
        auto_display: bool = False,
    ):
        """
        Create a new instance of AlertSelector.

        Parameters
        ----------
        alerts : pd.DataFrame
            DataFrame of alerts.
        action : Callable[..., None], optional
            Optional function to execute for each selected alert.
            (the default is None)
        columns : List[str], optional
            Override the default column names to use from `alerts`
            (the default is ['StartTimeUtc', 'AlertName',
            'CompromisedEntity', 'SystemAlertId'])
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self._w_output = widgets.Output(layout={"border": "1px solid black"})
        super().__init__(
            alerts=alerts, action=action, columns=columns, auto_display=auto_display
        )

    def display(self):
        """Display the interactive widgets."""
        self._select_top_alert()
        display(
            widgets.VBox([self._w_filter_alerts, self._w_select_alert, self._w_output])
        )

    def _run_action(self, change=None):
        del change
        self._w_output.clear_output()
        with self._w_output:
            self.alert_action(self.selected_alert)

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns
        -------
            dict(str, str) -- Dictionary of parameter names

        """
        return {"provider_alert_id": self.alert_id}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""figure_dimension - helps set the width and height properties of a figure for plotting."""
from functools import wraps
from typing import Any, Callable

from bokeh.plotting import figure

from .._version import VERSION
from ..common.utility import export

__version__ = VERSION
__author__ = "Claudiu Toma"


@export
def set_figure_size(fig: figure, width: int, height: int) -> figure:
    """Set the figure size.

    Args:
        fig (figure): initial figure
        width (int): width dimension
        height (int): height dimension

    Returns
    -------
        figure: figure with correct width and height

    """
    if hasattr(figure(), "height"):
        setattr(fig, "height", height)
    if hasattr(figure(), "plot_height"):
        setattr(fig, "plot_height", height)
    if hasattr(figure(), "width"):
        setattr(fig, "width", width)
    if hasattr(figure(), "plot_width"):
        setattr(fig, "plot_width", width)
    return fig


_BOKEH_3_FIG_PARAMS = {
    "plot_height": "height",
    "plot_width": "width",
}
_BOKEH_2_FIG_PARAMS = {val: key for key, val in _BOKEH_3_FIG_PARAMS.items()}


def bokeh_figure(func) -> Callable[[Any], Any]:
    """
    Wrap bokeh 'figure' function for version-agnostic param naming.

    Parameters
    ----------
    func : Callable
        The function to wrap

    Returns
    -------
    Callable :
        The wrapped figure function.

    """

    @wraps(func)
    def set_figure_size_params(*args, **kwargs):
        """Re-write parameters names for Bokeh version."""
        # pylint: disable=comparison-with-callable
        if func == figure:
            param_mapper = (
                _BOKEH_3_FIG_PARAMS
                if hasattr(func(), "height")
                else _BOKEH_2_FIG_PARAMS
            )

            func_kwargs = {
                param_mapper.get(param, param): value for param, value in kwargs.items()
            }
            return func(*args, **func_kwargs)
        return func(*args, **kwargs)

    return set_figure_size_params



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data provider sub-package."""
import importlib
from functools import singledispatch
from typing import Dict

from ..._version import VERSION
from ..core.query_defns import DataEnvironment

# flake8: noqa: F403
from .driver_base import DriverBase  # noqa: F401

__version__ = VERSION

_ENVIRONMENT_DRIVERS = {
    DataEnvironment.LogAnalytics: ("azure_monitor_driver", "AzureMonitorDriver"),
    DataEnvironment.SecurityGraph: ("security_graph_driver", "SecurityGraphDriver"),
    DataEnvironment.Kusto: ("azure_kusto_driver", "AzureKustoDriver"),
    DataEnvironment.MDATP: ("mdatp_driver", "MDATPDriver"),
    DataEnvironment.MDE: ("mdatp_driver", "MDATPDriver"),
    DataEnvironment.LocalData: ("local_data_driver", "LocalDataDriver"),
    DataEnvironment.OSQueryLogs: ("local_osquery_driver", "OSQueryLogDriver"),
    DataEnvironment.Splunk: ("splunk_driver", "SplunkDriver"),
    DataEnvironment.Mordor: ("mordor_driver", "MordorDriver"),
    DataEnvironment.Sumologic: ("sumologic_driver", "SumologicDriver"),
    DataEnvironment.ResourceGraph: ("resource_graph_driver", "ResourceGraphDriver"),
    DataEnvironment.M365D: ("mdatp_driver", "MDATPDriver"),
    DataEnvironment.Cybereason: ("cybereason_driver", "CybereasonDriver"),
    DataEnvironment.Elastic: ("elastic_driver", "ElasticDriver"),
    DataEnvironment.MSSentinel_New: ("azure_monitor_driver", "AzureMonitorDriver"),
    DataEnvironment.Kusto_New: ("azure_kusto_driver", "AzureKustoDriver"),
    DataEnvironment.VelociraptorLogs: (
        "local_velociraptor_driver",
        "VelociraptorLogDriver",
    ),
    DataEnvironment.MSSentinel_Legacy: ("kql_driver", "KqlDriver"),
    DataEnvironment.Kusto_Legacy: ("kusto_driver", "KustoDriver"),
    DataEnvironment.M365DGraph: ("mdatp_driver", "MDATPDriver"),
}

CUSTOM_PROVIDERS: Dict[str, type] = {}


@singledispatch
def import_driver(data_environment) -> type:
    """Unsupported type for environment."""
    raise TypeError(
        "'data_environment' must be a str or DataEnvironment type.",
        f"Called with type: {type(data_environment)}",
    )


@import_driver.register
def _(data_environment: DataEnvironment) -> type:
    """Import driver class for a data environment."""
    mod_name, cls_name = _ENVIRONMENT_DRIVERS.get(data_environment, (None, None))

    if not (mod_name and cls_name):
        raise ValueError(
            f"No driver available for environment {data_environment.name}.",
            "Possible values are:",
            ", ".join(env.name for env in _ENVIRONMENT_DRIVERS),
        )

    imp_module = importlib.import_module(
        f"msticpy.data.drivers.{mod_name}", package="msticpy"
    )
    return getattr(imp_module, cls_name)


@import_driver.register
def _(data_environment: str) -> type:
    """Import custom driver class for a data environment."""
    if plugin_cls := CUSTOM_PROVIDERS.get(data_environment):
        return plugin_cls

    raise ValueError(
        f"No driver available for environment {data_environment}.",
        "Possible values are:",
        ", ".join(CUSTOM_PROVIDERS),
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""OAuthApplication Entity class."""
from typing import Any, List, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class OAuthApplication(Entity):
    """
    OAuthApplication Entity class.

    Attributes
    ----------
    OAuthAppId : str
        Global ID of the application (in AAD, "Application ID" of the application object)
    OAuthObjectId : str
        Object ID of the service principal in AAD
    Name : str
        Name of the app
    TenantId : str
        AAD tenant ID in which the app was installed
    PublisherName : str
        The publisher name of the app
    Risk: str
        The app risk - like Low, Medium, High or Unknown
    Permissions: List[str]
        List of permissions that were requested by the application, and their severities
    RedirectURLs : List[str]
        List of redirect urls
    AuthorizedBy : int
        How many users consented the app

    """

    ID_PROPERTIES = ["OAuthAppId", "OAuthObjectId"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.OAuthAppId: Optional[str] = None
        self.OAuthObjectId: Optional[str] = None
        self.Name: Optional[str] = None
        self.TenantId: Optional[str] = None
        self.PublisherName: Optional[str] = None
        self.Risk: Optional[str] = None
        self.Permissions: List[str] = []
        self.RedirectURLs: List[str] = []
        self.AuthorizedBy: int = 0

        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self):
        """Return Entity Description."""
        return self.Name or self.OAuthAppId or self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Name or self.OAuthAppId or self.__class__.__name__

    _entity_schema = {
        # OAuthAppId (type System.String)
        "OAuthAppId": None,
        # OAuthObjectId (type System.String)
        "OAuthObjectId": None,
        # TenantId (type System.String)
        "TenantId": None,
        "PublisherName": None,
        "Risk": None,
        "Permissions": None,
        "RedirectURLs": None,
        "AuthorizedBy": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Common MSTICPy modules.

This is sub-package containing core utility, authentication and configuration
modules and classes.

Some of the modules and classes here:

- azure_auth - az_connect function
- pkg_config - the main settings management module
- secret_settings - handles Azure Key Vault and keyring-stored secrets
- provider_settings - allows access to settings that can redirect appropriate
  settings to secret_settings to retrieve from Key Vault/keyring
- wsconfig - Workspace config manager for MS Sentinel authentication
- utility and timespan

"""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Tree Visualization."""
import textwrap
from collections import Counter
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import pandas as pd

from .._version import VERSION
from .proc_tree_schema import ColNames as Col
from .proc_tree_schema import ProcSchema

__version__ = VERSION
__author__ = "Ian Hellen"


def get_process_key(procs: pd.DataFrame, source_index: int) -> str:
    """
    Return the process key of the process given its source_index.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events
    source_index : int, optional
        source_index of the process record

    Returns
    -------
    str
        The process key of the process.

    """
    return procs[procs[Col.source_index] == source_index].iloc[0].name  # type: ignore


# def build_process_key(  # type: ignore  # noqa: F821
#     source_proc: pd.Series,
#     schema: "ProcSchema"
# ) -> str:
#     """
#     Return a process key from a process event.

#     Parameters
#     ----------
#     source_proc : pd.Series, optional
#         Source process
#     schema : ProcSchema, optional
#         The data schema to use, by default None
#         - if None the schema will be inferred

#     Returns
#     -------
#     str
#         Process key of the process

#     """
#     if schema is None:
#         schema = infer_schema(source_proc)
#     proc_path = source_proc[schema.process_name].lower()
#     pid = source_proc[schema.process_id]
#     tstamp = pd.to_datetime(source_proc[schema.time_stamp]).strftime(TS_FMT_STRING)
#     return f"{proc_path}{pid}{tstamp}"


def get_roots(procs: pd.DataFrame) -> pd.DataFrame:
    """
    Return the process tree roots for the current data set.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)

    Returns
    -------
    pd.DataFrame
        Process Tree root processes

    """
    return procs[procs["IsRoot"]]


def get_process(procs: pd.DataFrame, source: Union[str, pd.Series]) -> pd.Series:
    """
    Return the process event as a Series.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    pd.Series
        Process row

    Raises
    ------
    ValueError
        If unknown type is supplied as `source`

    """
    if isinstance(source, str):
        return procs.loc[source]
    if isinstance(source, pd.Series):
        return source
    raise ValueError("Unknown type for source parameter.")


def get_parent(
    procs: pd.DataFrame, source: Union[str, pd.Series]
) -> Optional[pd.Series]:
    """
    Return the parent of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    Optional[pd.Series]
        Parent Process row or None if no parent was found.

    """
    proc = get_process(procs, source)
    if proc.parent_key in procs.index:
        return procs.loc[proc.parent_key]
    return None


def get_root(procs: pd.DataFrame, source: Union[str, pd.Series]) -> pd.Series:
    """
    Return the root process for the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    pd.Series
        Root process

    """
    proc = get_process(procs, source)
    p_path = proc.path.split("/")
    root_proc = procs[procs[Col.source_index] == p_path[0]]
    return root_proc.iloc[0]


def get_root_tree(procs: pd.DataFrame, source: Union[str, pd.Series]) -> pd.DataFrame:
    """
    Return the process tree to which the source process belongs.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    pd.DataFrame
        Process Tree

    """
    proc = get_process(procs, source)
    p_path = proc.path.split("/")
    return procs[procs["path"].str.startswith(p_path[0])]


def get_tree_depth(procs: pd.DataFrame) -> int:
    """
    Return the depth of the process tree.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)

    Returns
    -------
    int
        Tree depth

    """
    return procs["path"].str.count("/").max() + 1


def get_children(
    procs: pd.DataFrame, source: Union[str, pd.Series], include_source: bool = True
) -> pd.DataFrame:
    """
    Return the child processes for the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        If True include the source process in the results, by default True

    Returns
    -------
    pd.DataFrame
        Child processes

    """
    proc = get_process(procs, source)
    current_index_name = procs.index.name
    children = procs[procs[Col.parent_key] == proc.name]
    if include_source:
        proc_df = _fill_empty_columns(pd.DataFrame(proc).T)
        children = pd.concat([proc_df, _fill_empty_columns(children)])
        children.index.name = current_index_name
    return children


def get_descendents(
    procs: pd.DataFrame,
    source: Union[str, pd.Series],
    include_source: bool = True,
    max_levels: int = -1,
) -> pd.DataFrame:
    """
    Return the descendents of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        Include the source process in the results, by default True
    max_levels : int, optional
        Maximum number of levels to descend, by default -1 (all levels)

    Returns
    -------
    pd.DataFrame
        Descendent processes

    """
    proc = get_process(procs, source)

    descendents = []
    parent_keys = [proc.name]
    level = 0
    current_index_name = procs.index.name
    rem_procs: Optional[pd.DataFrame] = None
    while max_levels == -1 or level < max_levels:
        if rem_procs is not None:
            # pylint: disable=unsubscriptable-object
            children = rem_procs[rem_procs[Col.parent_key].isin(parent_keys)]
            rem_procs = rem_procs[~rem_procs[Col.parent_key].isin(parent_keys)]
            # pylint: enable=unsubscriptable-object
        else:
            children = procs[procs[Col.parent_key].isin(parent_keys)]
            rem_procs = procs[~procs[Col.parent_key].isin(parent_keys)]
        if children.empty:
            break
        descendents.append(children)
        parent_keys = children.index  # type: ignore
        level += 1

    if descendents:
        desc_procs = pd.concat(descendents)
    else:
        desc_procs = pd.DataFrame(columns=proc.index, index=None)
        desc_procs.index.name = Col.proc_key

    desc_procs = _fill_empty_columns(desc_procs)
    if include_source:
        proc_df = _fill_empty_columns(pd.DataFrame(proc).T)
        desc_procs = pd.concat([proc_df, desc_procs])
        desc_procs.index.name = current_index_name
    return desc_procs.sort_values("path")


def get_ancestors(procs: pd.DataFrame, source, include_source=True) -> pd.DataFrame:
    """
    Return the ancestor processes of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        Include the source process in the results, by default True

    Returns
    -------
    pd.DataFrame
        Ancestor processes

    """
    proc = get_process(procs, source)
    p_path = proc.path.split("/")
    if not include_source:
        p_path.remove(proc.source_index)
    return procs[procs[Col.source_index].isin(p_path)].sort_values("path")


def get_siblings(
    procs: pd.DataFrame, source: Union[str, pd.Series], include_source: bool = True
) -> pd.DataFrame:
    """
    Return the processes that share the parent of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        Include the source process in the results, by default True

    Returns
    -------
    pd.DataFrame
        Sibling processes.

    """
    parent = get_parent(procs, source)
    proc = get_process(procs, source)
    siblings = get_children(procs, parent, include_source=False)  # type: ignore
    if not include_source:
        return siblings[siblings.index != proc.name]
    return siblings


def get_summary_info(procs: pd.DataFrame) -> Dict[str, int]:
    """
    Return summary information about the process trees.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)

    Returns
    -------
    Dict[str, int]
        Summary statistic about the process tree

    """
    summary: Dict[str, Any] = {}
    summary["Processes"] = len(procs)
    summary["RootProcesses"] = len(procs[procs["IsRoot"]])
    summary["LeafProcesses"] = len(procs[procs["IsLeaf"]])
    summary["BranchProcesses"] = len(procs[procs["IsBranch"]])
    summary["IsolatedProcesses"] = len(procs[(procs["IsRoot"]) & (procs["IsLeaf"])])
    summary["LargestTreeDepth"] = procs["path"].str.count("/").max() + 1
    return summary


class TemplateLine(NamedTuple):
    """
    Template definition for a line in text process tree.

    Notes
    -----
    The items attribute must be a list of tuples, where each
    tuple is (<display_name>, <column_name>).

    """

    items: List[Tuple[str, str]] = []
    wrap: int = 80


def tree_to_text(
    procs: pd.DataFrame,
    schema: Optional[Union[ProcSchema, Dict[str, str]]] = None,
    template: Optional[List[TemplateLine]] = None,
    sort_column: str = "path",
    wrap_column: int = 0,
) -> str:
    """
    Return text rendering of process tree.

    Parameters
    ----------
    procs : pd.DataFrame
        The process tree DataFrame.
    schema : Optional[Union[ProcSchema, Dict[str, str]]], optional
        The schema to use for mapping the DataFrame column
        names, by default None
    template : Optional[List[TemplateLine]], optional
        A manually created template to use to create the node
        formatting, by default None
    sort_column : str, optional
        The column to sort the DataFrame by, by default "path"
    wrap_column : int, optional
        Override any template-specified wrap limit, by default 0

    Returns
    -------
    str
        The formatted process tree string.

    Raises
    ------
    ValueError
        If neither of
    """
    if not schema and not template:
        raise ValueError(
            "One of 'schema' and 'template' must be supplied", "as parameters."
        )
    template = template or _create_proctree_template(schema)  # type: ignore
    output: List[str] = []
    for _, row in procs.sort_values(sort_column).iterrows():
        depth_count = Counter(row.path).get("/", 0)
        header = _node_header(depth_count)

        # handle first row separately since it needs a header
        tmplt_line = template[0]
        out_line = "  ".join(
            f"{name}: {row[col]}" if name else f"{row[col]}"
            for name, col in tmplt_line.items
        )
        indent = " " * len(header) + " "
        out_line = "\n".join(
            textwrap.wrap(
                out_line,
                width=wrap_column or tmplt_line.wrap,
                subsequent_indent=indent,
            )
        )
        output.append(f"{header} {out_line}\n")

        # process subsequent rows
        for tmplt_line in template[1:]:
            out_line = "  ".join(
                f"{name}: {row[col]}" for name, col in tmplt_line.items
            )
            out_line = "\n".join(
                textwrap.wrap(
                    out_line,
                    width=wrap_column or tmplt_line.wrap,
                    initial_indent=indent,
                    subsequent_indent=indent + "   ",
                )
            )
            output.extend([out_line, "\n"])

    return "".join(output)


def _create_proctree_template(
    schema: Union[ProcSchema, Dict[str, str]]
) -> List[TemplateLine]:
    """Create a template from the schema."""
    if isinstance(schema, dict):
        schema = ProcSchema(**schema)
    template_lines: List[TemplateLine] = [
        TemplateLine(
            items=[("Process", schema.process_name), ("PID", schema.process_id)]
        ),
        TemplateLine(items=[("Time", schema.time_stamp)]),
    ]
    if schema.cmd_line:
        template_lines.append(TemplateLine(items=[("Cmdline", schema.cmd_line)]))
    acct_items = []
    if schema.user_id:
        acct_items.append(("Account", schema.user_id))
    if schema.logon_id:
        acct_items.append(("Account", schema.logon_id))
    if acct_items:
        template_lines.append(TemplateLine(items=acct_items))
    return template_lines


def _node_header(depth_count):
    """Return text tree node header given tree depth."""
    return "+ " if depth_count == 0 else "   " * depth_count + "+-- "


def _fill_empty_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Fill empty columns in the DataFrame."""
    df = df.copy()
    for col in df.columns[df.isna().all()]:
        if df[col].dtype == "object":
            df[col] = df[col].fillna("")
        elif pd.api.types.is_numeric_dtype(df[col].dtype):
            df[col] = df[col].fillna(0)
    return df



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MDATP OData Driver class."""
from typing import Any, Optional, Union

import pandas as pd

from ..._version import VERSION
from ...auth.azure_auth_core import AzureCloudConfig
from ...auth.cloud_mappings import (
    get_defender_endpoint,
    get_m365d_endpoint,
    get_m365d_login_endpoint,
)
from ...common.data_utils import ensure_df_datetimes
from ...common.utility import export
from ..core.query_defns import DataEnvironment
from .odata_driver import OData, QuerySource, _get_driver_settings

__version__ = VERSION
__author__ = "Pete Bryan"


@export
class MDATPDriver(OData):
    """KqlDriver class to retrieve date from MS Defender APIs."""

    CONFIG_NAME = "MicrosoftDefender"
    _ALT_CONFIG_NAMES = ["MDATPApp"]

    def __init__(
        self, connection_str: Optional[str] = None, instance: str = "Default", **kwargs
    ):
        """
        Instantiate MSDefenderDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string
        instance : str, optional
            The instance name from config to use

        """
        super().__init__(**kwargs)
        cs_dict = _get_driver_settings(
            self.CONFIG_NAME, self._ALT_CONFIG_NAMES, instance
        )

        self.cloud = cs_dict.pop("cloud", "global")
        if "cloud" in kwargs and kwargs["cloud"]:
            self.cloud = kwargs["cloud"]

        api_uri, oauth_uri, api_suffix = _select_api_uris(
            self.data_environment, self.cloud
        )
        self.add_query_filter(
            "data_environments", ("MDE", "M365D", "MDATP", "GraphHunting")
        )

        self.req_body = {
            "client_id": None,
            "client_secret": None,
            "grant_type": "client_credentials",
            "resource": api_uri,
        }
        self.oauth_url = oauth_uri
        self.api_root = api_uri
        self.api_ver = "api"
        self.api_suffix = api_suffix
        if self.data_environment == DataEnvironment.M365D:
            self.scopes = [f"{api_uri}/AdvancedHunting.Read"]
        elif self.data_environment == DataEnvironment.M365DGraph:
            self.api_ver = kwargs.get("api_ver", "v1.0")
            self.req_body = {
                "client_id": None,
                "client_secret": None,
                "grant_type": "client_credentials",
                "scope": f"{self.api_root}.default",
            }
            self.scopes = [f"{api_uri}/ThreatHunting.Read.All"]
        else:
            self.scopes = [f"{api_uri}/AdvancedQuery.Read"]

        if connection_str:
            self.current_connection = connection_str
            self.connect(connection_str)

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        del query_source, kwargs
        data, response = self.query_with_results(
            query, body=True, api_end=self.api_suffix
        )
        if isinstance(data, pd.DataFrame):
            # If we got a schema we should convert the DateTimes to pandas datetimes
            if ("Schema" or "schema") not in response:
                return data

            if self.data_environment == DataEnvironment.M365DGraph:
                date_fields = [
                    field["name"]
                    for field in response["schema"]
                    if field["type"] == "DateTime"
                ]
            else:
                date_fields = [
                    field["Name"]
                    for field in response["Schema"]
                    if field["Type"] == "DateTime"
                ]
            data = ensure_df_datetimes(data, columns=date_fields)
            return data
        return response


def _select_api_uris(data_environment, cloud):
    """Return API and login URIs for selected provider type."""
    login_uri = get_m365d_login_endpoint(cloud)
    if data_environment == DataEnvironment.M365D:
        return (
            get_m365d_endpoint(cloud),
            f"{login_uri}{{tenantId}}/oauth2/token",
            "/advancedhunting/run",
        )
    if data_environment == DataEnvironment.M365DGraph:
        az_cloud_config = AzureCloudConfig(cloud=cloud)
        api_uri = az_cloud_config.endpoints.get("microsoftGraphResourceId")
        graph_login = az_cloud_config.authority_uri
        return (
            api_uri,
            f"{graph_login}{{tenantId}}/oauth2/v2.0/token",
            "/security/runHuntingQuery",
        )
    return (
        get_defender_endpoint(cloud),
        f"{login_uri}{{tenantId}}/oauth2/token",
        "/advancedqueries/run",
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Deprecated - module auditdextract.py has moved.

See :py:mod:`msticpy.transform.auditdextract`
"""
import warnings

from .._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"


# flake8: noqa: F403, F401
# pylint: disable=wildcard-import, unused-wildcard-import, unused-import
from ..transform.auditdextract import *

WARN_MSSG = (
    "This module has moved to msticpy.transform.auditdextract\n"
    "Please change your import to reflect this new location."
    "This will be removed in MSTICPy v2.2.0"
)
warnings.warn(WARN_MSSG, category=DeprecationWarning)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data storage sub-package."""

from ..._version import VERSION

# pylint: disable=unused-import
# flake8: noqa: F403
from .azure_blob_storage import AzureBlobStorage  # noqa: F401

__version__ = VERSION
